;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "ob-athena" "2.2.5"
  "Run AWS Athena queries from Org Babel."
  '((emacs "26.1"))
  :url "https://github.com/will-abb/aws-athena-babel"
  :commit "132ccc35c13e772ed39813d949d0089ea0c66a66"
  :revdesc "v2.2.5-0-g132ccc35c13e"
  :keywords '("aws" "athena" "org" "babel" "sql" "tools")
  :authors '(("Williams Bosch-Bello" . "williamsbosch@gmail.com"))
  :maintainers '(("Williams Bosch-Bello" . "williamsbosch@gmail.com")))
